###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2001  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: checkpass.tcl,v 1.3 2001/06/07 08:19:26 tothwolf Exp $
##

# name: checkpass
# version: 1.12.0.0
# config: checkpass.conf
# description: Password checking utility that adds a dcc command called \
#	`.checkpass', performs password checking on join for defined \
#	channels, and allows the deletion of users after a set number of \
#	days if they have not set a password.
# requires: wpLog commandChar putnot daysLeft grammarPlural randstring \
#	validHandle findChanFlag findHandOnChannels convertSeconds \
#	moveFields

namespace eval checkpass {

  ##
  ## Check for users with empty passwords: dcc
  ##
  ## Args:    handle, idx, arg
  ## Returns: 1
  ##
  proc dcc:checkpass {hand idx arg} {
    set userCount 0
    set botCount 0
    foreach user [userlist] {
      if {[passwdok $user ""]} then {
        if {![matchattr $user b]} then {
          lappend userList $user,
          incr userCount
        } else {
          lappend botList $user,
          incr botCount
        }
      }
    }
    if {$userCount} then {
      putdcc $idx "Users without a password ($userCount): [string trimright [join $userList] ,]"
    } else {
      putdcc $idx "All users have a password set."
    }
    if {$botCount} then {
      putdcc $idx "Bots without a password ($botCount): [string trimright [join $botList] ,]"
    } else {
      putdcc $idx "All bots have a password set."
    }
    return 1
  }

  ##
  ## Check for users with empty passwords when they join channels
  ##
  ## Args:    nick, user@host, handle, channel
  ## Returns: 0
  ##
  proc join:checkpass {nick uhost hand chan} {
    variable NamespaceCurrent
    variable configData
    variable CheckUsers

    if {([validHandle $hand]) && ([passwdok $hand ""])} then {
      if {($configData(dkrandpass)) && \
          (([matchattr $hand dk|dk $chan]) || \
           ([findChanFlag $hand dk]))} then {
        setuser $hand PASS [randstring 8]
      } else {
        if {![info exists CheckUsers($hand)]} then {
          set CheckUsers($hand) [clock seconds]
          wpLog o * $NamespaceCurrent "$nick!$uhost ($hand) does not have a password set, sending a notice..."
          utimer 5 "${NamespaceCurrent}::SendNotice [split $nick] [split $hand]"
        }
      }
    }
    return 0
  }

  ##
  ## Remove from CheckUsers if user signs off
  ##
  ## Args:    nick, user@host, handle, channel, arg
  ## Returns: 0
  ##
  proc sign:checkpass {nick uhost hand chan arg} {
    RemoveFromCheckUsers $hand
    return 0
  }

  ##
  ## Remove from CheckUsers if user parts a chan and are not on other channels
  ##
  ## Args:    nick, user@host, handle, channel
  ## Returns: 0
  ##
  proc part:checkpass {nick uhost hand chan {arg ""}} {
    RemoveFromCheckUsers $hand 1
    return 0
  }

  ##
  ## Remove from CheckUsers if user signs off due to a netsplit
  ##
  ## Args:    nick, user@host, handle, channel
  ## Returns: 0
  ##
  proc splt:checkpass {nick uhost hand chan} {
    RemoveFromCheckUsers $hand
    return 0
  }

  ##
  ## Change data in CheckUsers if user changes nicks
  ##
  ## Args:    nick, user@host, handle, channel, newnick
  ## Returns: 0
  ##
  proc nick:checkpass {nick uhost hand chan newnick} {
    variable CheckUser

    if {([validuser $hand]) && ([info exists CheckUser($hand)])} then {
      set CheckUser($hand) [lreplace $CheckUser($hand) 1 1 [split $newnick]]
    }
    return 0
  }

  ##
  ## Annoy people who refuse to set a password
  ##
  ## Args:    minute, hour, day, month, year
  ## Returns: 0
  ##
  proc time:minutely {minute hour day month year} {
    variable configData
    variable CheckUsers

    if {[string compare "" [set users [array get CheckUsers]]]} then {
      set time [clock seconds]
      set delay [expr [llength [split $users]] * 60]
      foreach {hand seconds} $users {
        if {[passwdok $hand ""]} then {
          if {[expr $seconds - $time] >= $delay} then {
            foreach chan [channels] {
              if {[handonchan $hand $chan]} then {
                set CheckUsers($hand) $time
                SendNotice [hand2nick $nick $chan] $hand
                return 0
              }
            }
          }
        }
      } else {
        unset CheckUsers($hand)
      }
    }
    return 0
  }

  ##
  ## Unset handles in array who have set a password
  ##
  ## Args:    minute, hour, day, month, year
  ## Returns: 0
  ##
  proc time:hourly {minute hour day month year} {
    variable CheckUsers

    foreach {hand seconds} [array get CheckUsers] {
      if {(![passwdok $hand ""]) || (![findHandOnChannels $hand])} then {
        unset CheckUsers($hand)
      }
    }
    return 0
  }

  ##
  ## Delete users who still don't have a password set
  ##
  ## Args:    minute, hour, day, month, year
  ## Returns: 0
  ##
  proc time:daily {minute hour day month year} {
    variable NamespaceCurrent
    variable configData

    set time [clock seconds]
    foreach user [userlist -b] {
      set passtime [getuser $user XTRA $configData(timefield)]
      if {($configData(passtime)) && ([passwdok $user ""])} then {
        if {![string compare "" $passtime]} then {
          setuser $user XTRA [string toupper $configData(timefield)] $time
        } elseif {[expr $time - $passtime] >= [convertSeconds from days $configData(passtime)]} then {
          wpLog o * $NamespaceCurrent "Deleted user: $user, no password was set after [grammarPlural $configData(passtime) day]."
          deluser $user
        }
      } elseif {[string compare "" $passtime]} then {
        setuser $user XTRA $configData(timefield)
      }
    }
    return 0
  }

  ##
  ## Remove data for the given handle from CheckUsers
  ##
  ## Args:    handle, check channels {0|1}
  ## Returns:
  ##
  proc RemoveFromCheckUsers {hand {checkchans 0}} {
    variable CheckUsers

    if {([validuser $hand]) && ([info exists CheckUsers($hand)]) && \
        ((!$checkchans) || (![findHandOnChannels $hand]))} then {
      unset CheckUsers($hand)
    }
    return
  }

  ##
  ## Send a notice about setting a password to the given nick
  ##
  ## Args:    nickname, handle
  ## Returns: nothing
  ##
  proc SendNotice {nick hand} {
    variable configData

    set cmdchar [commandChar msg]
    putnot $nick "You do not have a password set."
    if {$configData(passtime)} then {
      putnot $nick "You access will be removed in [grammarPlural [ListPassTime $hand] day] unless you set one."
    }
    putnot $nick "Please set one now, by typing: ${cmdchar}pass <your password>"
    putnot $nick "For more details: ${cmdchar}help pass"
    return
  }

  ##
  ## List password time for a user
  ##
  ## Args:    user
  ## Returns: password time
  ##
  proc ListPassTime {user} {
    variable NamespaceCurrent
    variable configData

    if {$configData(passtime) && [passwdok $user ""]} then {
      if {![string compare "" [set passtime [getuser $user XTRA $configData(timefield)]]]} then {
        set passtime [clock seconds]
        setuser $user XTRA [string toupper $configData(timefield)] $passtime
      }
      return [daysLeft $passtime $configData(passtime)]
    }
    return
  }

  ##
  ## Trace configuration changes
  ##
  ## Args:    name1, name2, operation
  ## Returns: nothing
  ##
  proc TraceConfigData {name1 name2 op} {
    variable configData
    variable oldConfigData
    variable DiceConfig

    if {[string compare $oldConfigData($name2) $configData($name2)]} then {
      switch -exact -- $name2 {
        channels {
          BindTable unbind channels
          set oldConfigData(channels) $configData(channels)
          BindTable bind channels
        }
        timefield {
          moveFields $oldConfigData(timefield) [string toupper $configData(timefield)]
          set oldConfigData(timefield) $configData(timefield)
        }
      }
    }
    return
  }

  ##
  ## Trace and respond to variable changes
  ##
  ## Args:    action {variable|vdelete}
  ## Returns: nothing
  ##
  proc TraceTable {action} {
    variable NamespaceCurrent
    variable configData

    trace $action configData w ${NamespaceCurrent}::TraceConfigData
    return
  }

  ##
  ## Init module specific things at load time
  ##
  ## Args:    action {load|unload}
  ## Returns: nothing
  ##
  proc InitTable {action} {
    variable NamespaceCurrent
    variable configData
    variable oldConfigData

    switch -exact -- $action {
      load {
        if {![info exists NamespaceCurrent]} then {
          set NamespaceCurrent [namespace current]
        }
        foreach {name data} [array get configData] {
          if {![info exists oldConfigData($name)]} then {
            set oldConfigData($name) $data
          }
        }
      }
    }
    return
  }

  ##
  ## Bind procs to commands and events
  ##
  ## Args:    action {bind|unbind}, what {type|*}, args {...}
  ## Returns: nothing
  ##
  proc BindTable {action what args} {
    variable NamespaceCurrent
    variable oldConfigData

    if {![string compare * $what]} then {
      set what "dcc time channels"
    }
    foreach type $what {
      switch -exact -- $type {
        dcc {
          $action dcc m|m checkpass ${NamespaceCurrent}::dcc:checkpass
        }
        time {
          $action time -|- "* * * * *" ${NamespaceCurrent}::time:minutely
          $action time -|- "00 * * * *" ${NamespaceCurrent}::time:hourly
          $action time -|- "00 00 * * *" ${NamespaceCurrent}::time:daily
        }
        channels {
          if {([info exists oldConfigData(channels)]) && \
              ([string compare "" $oldConfigData(channels)]) && \
              ([string compare - $oldConfigData(channels)])} then {
            foreach chan $oldConfigData(channels) {
              $action join -|- "$chan *" ${NamespaceCurrent}::join:checkpass
              $action sign -|- "$chan *" ${NamespaceCurrent}::sign:checkpass
              $action part -|- "$chan *" ${NamespaceCurrent}::part:checkpass
              $action nick -|- "$chan *" ${NamespaceCurrent}::nick:checkpass
              $action splt -|- "$chan *" ${NamespaceCurrent}::splt:checkpass
            }
          }
        }
      }
    }
    return
  }

  # Configuration data defaults
  array set configDataDefaults {
    channels -
    dkrandpass 1
    passtime 10
    timefield passtime
  }

  # Configuration data descriptions
  array set configDataDesc {
    channels "\[channels\] channels to check on, '#chan1 #chan2', '*' - all, \"\" - none"
    dkrandpass "\[0/1\] set a random password for a +d or +k user?"
    passtime "\[days\] max number of days before a user with an empty password is deleted"
    timefield "\[field\] field for storing check timestamp"
  }

} ;# namespace checkpass

return
