###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2001  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: texttools.tcl,v 1.4 2001/06/01 05:38:49 tothwolf Exp $
##

# name: texttools
# version: 1.0.0.0
# description: Text formatting functions for use by other modules.
# provides: padList randstring randomIndex splitNumber isNumber \
#	convertFileSize grammarPlural grammarA grammarWas b v u o c ansi

namespace eval texttools {

  ##
  ## Format the given text into padded columns of a certain width
  ##
  ## Args:    text, column width, pad size
  ## Returns: formatted list
  ##
  proc padList {text columnWidth padLength} {
    # Sanity check columnWidth and padLength
    if {($columnWidth >= 1) && ($padLength >= 1)} then {
      # Primary for loop; once per output line
      for {
        set listLength [llength $text]	;# Total length of list
        set listIndex 0			;# Current index of list
        set subListIndex 0		;# Starting index of Secondary for loop
      } {$listIndex < $listLength} {
        incr listIndex $columnWidth
      } {
        # Secondary for loop; once per word
        for {
          set line ""			;# Finished line
          set columnIndex 0		;# Current index of line
        } {$columnIndex < $columnWidth} {
          incr columnIndex
          incr subListIndex
        } {
          # Word from list
          set word [lindex $text $subListIndex]
          # Don't pad word if we are at max width or if there are no more words
          if {([expr $columnIndex + 1] == $columnWidth) || \
              (![string compare "" [lindex $text [expr $subListIndex + 1]]])} then {
            append line $word
            incr subListIndex		;# Do this here since we break the loop
            break
          } else {
            # Pad word to fit
            append line [format "%-${padLength}s" $word]
          }
        }
        # Append finished line to output list
        if {[string compare "" $line]} then {
          lappend list $line
        }
      }
      # Whew...that was alot of work!
      if {[info exists list]} then {
        return $list
      }
    }
    return
  }

  ##
  ## Return a random string of the given length
  ##
  ## Args:    length, chars
  ## Returns: random string
  ##
  proc randstring {length {chars abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789}} {
    set count [string length $chars]
    for {set index 0} {$index < $length} {incr index} {
      append result [string index $chars [rand $count]]
    }
    return $result
  }

  ##
  ## Return random data from the given list
  ##
  ## Args:    list
  ## Returns: random data from list
  ##
  proc randomIndex {list} {
    set list [split $list]
    set length [llength $list]
    if {$length} then {
      return [lindex $list [rand $length]]
    }
    return
  }

  ##
  ## Split a leading number from text for each element in a list
  ##
  ## Args:    list
  ## Returns: list
  ##
  proc splitNumber {list} {
    set ret ""
    foreach element $list {
      set length [string length $element]
      set count 0
      set string ""
      set word ""
      while {$count < $length} {
        if {[regexp \[0-9\] [string index $element $count]]} then {
          if {[regexp \[0-9\] [string index $element [expr $count + 1]]]} then {
            append string [string index $element $count]
          } else {
            append string [string index $element $count] " "
          }
        } else {
          set tmp [string index $element [expr $count + 1]]
          if {(![regexp \[0-9\] $tmp]) && ([string compare "" $tmp])} then {
            append string [string index $element $count]
          } else {
            append string [string index $element $count]
            lappend word $string
            set string ""
          }
        }
        incr count
      }
      if {[string compare "" $word]} then {
        lappend ret $word
      }
    }
    return $ret
  }

  ##
  ## Check if the given string is a valid number
  ##
  ## Args:    string
  ## Returns: 1 if the given string is a number,
  ##          0 otherwise
  ##
  proc isNumber {string} {
    if {[string compare $string ""]} then {
      # Remove leading hyphen
      if {[regexp -- ^- $string]} then {
        regsub -- ^- $string "" string
      }
      # Remove first decimal point
      if {[regexp -- \\. $string]} then {
        regsub -- \\. $string "" string
      }
      # Valid number?
      if {![regexp -- \[^0-9\] $string]} then {
        return 1
      }
    }
    return 0
  }

  ##
  ## Check if the given number is exactly divisible by the given divisor
  ##
  ## Args:   number, divisor
  ## Returns: 1 if the number is exactly divisible by the divisor
  ##          0 otherwise, or if a divide by 0 error occurs
  ##
  proc isDivisible {number divisor} {
    if {[catch {
      if {[expr (($number + 0.0) / $divisor) == round($number / $divisor)]} then {
        return 1
      } else {
        return 0
      }
    }]} then {
      return 0
    }
  }

  ##
  ## Convert bytes to kb or mb
  ##
  ## Args:    bytes
  ## Returns: kb or mb
  ##
  proc convertFileSize {size} {
    if {$size >= 1024000} then {
      return [expr round($size / 1024000.0)]Mb
    } elseif {$size >= 1024} then {
      return [expr round($size / 1024.0)]Kb
    } else {
      return ${size}b
    }
  }

  ##
  ## Fix 's' grammar for the given number and text
  ##
  ## Args:   number, text
  ## Returns: "number text(s)"
  ##
  proc grammarPlural {number text} {
    if {$number != 1} then {
      return "$number ${text}s"
    }
    return "$number $text"
  }

  ##
  ## Fix 'a/an' grammar for the given case and text
  ##
  ## Args:    text, uppercase {0/1}
  ## Returns: "a/an text"
  ##
  proc grammarA {text {uppercase 0}} {
    if {$uppercase} then {
      set a A
    } else {
      set a a
    }
    if {[regexp -- "^\[aeiouAEIOU\]" $text]} then {
      return "${a}n $text"
    }
    return "$a $text"
  }

  ##
  ## Fix 'was/were' grammar for the given text
  ##
  ## Args:    text
  ## Returns: "was/were text"
  ##
  proc grammarWas {text} {
    if {![string compare s \
          [string index $text [expr [string length $text] - 1]]]} then {
      return "$text were"
    } else {
      return "$text was"
    }
  }

  ##
  ## Return bold code for irc text formatting
  ##
  ## Args:    none
  ## Returns: ^B bold code
  ##
  proc b {} {
    return "\002"
  }

  ##
  ## Return inverse code for irc text formatting
  ##
  ## Args:    none
  ## Returns: ^V inverse code
  ##
  proc v {} {
    return "\026"
  }

  ##
  ## Return underline code for irc text formatting
  ##
  ## Args:    none
  ## Returns: ^_ underline code
  ##
  proc u {} {
    return "\037"
  }

  ##
  ## Return reset code for irc text formatting
  ##
  ## Args:    none
  ## Returns: ^O reset code
  ##
  proc o {} {
    return "\017"
  }

  ##
  ## Return mIRC color codes
  ##
  ## Args:   color number
  ## Returns: ^C<color number>
  ##
  # color char
  proc c {args} {
    return "\003[join $args]"
  }

  ##
  ## Return ansi color and formatting codes
  ##
  ## Args:    color/mode
  ## Returns: ansi escape code
  ##
  proc ansi {color} {
    set esc [format %c 27]
    switch -exact -- $color {
      norm {
        return "$esc\[0m"
      }
      black {
        return "$esc\[0;30m"
      }
      red {
        return "$esc\[0;31m"
      }
      green {
        return "$esc\[0;32m"
      }
      yellow {
        return "$esc\[0;33m"
      }
      blue {
        return "$esc\[0;34m"
      }
      purple {
        return "$esc\[0;35m"
      }
      cyan {
        return "$esc\[0;36m"
      }
      white {
        return "$esc\[0;37m"
      }
      bblack {
        return "$esc\[1;30m"
      }
      bred {
        return "$esc\[1;31m"
      }
      bgreen {
        return "$esc\[1;32m"
      }
      byellow {
        return "$esc\[1;33m"
      }
      bblue {
        return "$esc\[1;34m"
      }
      bpurple {
        return "$esc\[1;35m"
      }
      bcyan {
        return "$esc\[1;36m"
      }
      bwhite {
        return "$esc\[1;37m"
      }
      bold {
        return "$esc\[1m"
      }
      uline {
        return "$esc\[4m"
      }
      blink {
        return "$esc\[5m"
      }
      reverse {
        return "$esc\[7m"
      }
    }
  }

} ;# namespace texttools

return
