###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2001  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: timetools.tcl,v 1.2 2001/06/01 05:38:49 tothwolf Exp $
##

# name: timetools
# version: 1.0.0.0
# config: timetools.conf
# description: Time formatting functions for use by other modules.
# provides: timerExists timerId convertTime formatTime daysLeft \
#	convertSeconds convertDHM countClicksPerSecond \
#	averageClicksPerSecond calibrateClicksPerSecond clicksPerSecond
# requires: wpLog

namespace eval timetools {

  ##
  ## Check if a given command has a timer pending
  ##
  ## Args:    matchcmd, timercmd {timers|utimers}
  ## Returns: 1 if the given matchcmd has a pending timer
  ##          0 otherwise
  ##
  proc timerExists {matchcmd {timercmd timers}} {
    foreach timer [$timercmd] {
      foreach {seconds command id} $timer {break}
      if {![string compare $matchcmd $command]} then {
        return 1
      }
    }
    return 0
  }

  ##
  ## Find timerid for a given command
  ##
  ## Args:    matchcmd, timercmd {timers|utimers}
  ## Returns: timerid if the given command has a pending timer
  ##          nothing otherwise
  ##
  proc timerId {matchcmd {timercmd timers}} {
    foreach timer [$timercmd] {
      foreach {seconds command id} $timer {break}
      if {![string compare $matchcmd $command]} then {
        return $id
      }
    }
    return
  }

  ##
  ## Convert a unixtime timestamp to readable format
  ##
  ## Args:    unixtime timestamp
  ## Returns: 'HH:MM:SS' if less then 1 day ago,
  ##          'MMM DD YYYY' if more then 1 day ago
  ##
  proc convertTime {time} {
    if {[expr ([clock seconds] - $time) > 86400]} then {
      set reply [clock format $time -format "%b %d %Y"]
    } else {
      set reply [clock format $time -format "%H:%M:%S"]
    }
    return $reply
  }

  ##
  ## Convert a unixtime timestamp to 'MMM DD YYYY HH:MM:SS' format
  ##
  ## Args:    unixtime timestamp
  ## Returns: 'MMM DD YYYY HH:MM:SS'
  ##
  proc formatTime {time} {
    return [clock format $time -format "%b %d %Y %H:%M:%S"]
  }

  ##
  ## Calculate days remaining from the unixtime timestamp and day count given
  ##
  ## Args:    unixtime timestamp, days
  ## Returns: days remaining
  ##
  proc daysLeft {time days} {
    set seconds [expr [clock seconds] - $time]
    while {$seconds >= 86400} {
      incr seconds -86400
      incr days -1
    }
    if {$days > 0} then {
      return $days
    }
    return 0
  }

  ##
  ## Convert a value to/from seconds
  ##
  ## Args:    mode {from,to}, what, value
  ## Returns: converted value
  ##
  proc convertSeconds {mode what value} {
    switch -exact -- $mode {
      from {
        set expr *
      }
      to {
        set expr /
      }
      default {
        return
      }
    }
    switch -exact -- $what {
      seconds {
        return $value
      }
      minutes {
        return [expr $value $expr 60]
      }
      hours {
        return [expr $value $expr 3600]
      }
      days {
        return [expr $value $expr 86400]
      }
      weeks {
        return [expr $value $expr 604800]
      }
      months {
        return [expr $value $expr 2628000]
      }
      years {
        return [expr $value $expr 31536000]
      }
    }
    return
  }

  ##
  ## Converts XdXhXm time description to number of seconds
  ##
  ## Arg:     string
  ## Returns: seconds
  ##
  proc convertDHM {string} {
    set seconds 0
    foreach {type number} [splitNumber $string] {
      switch -exact -- [string tolower $type] {
        d {
          incr seconds [convertSeconds from days $number]
        }
        h {
          incr seconds [convertSeconds from hours $number]
        }
        m {
          incr seconds [convertSeconds from minutes $number]
        }
      }
    }
    return $seconds
  }

  ##
  ## Count number of click clicks in 1 second
  ##
  ## Args:    none
  ## Returns: number of click clicks in 1 second
  ##
  proc countClicksPerSecond {} {
    set clicks [clock clicks]
    after 1000
    return [expr [clock clicks] - $clicks]
  }

  ##
  ## Average number of clock clicks per second by number of rounds
  ##
  ## Args:    rounds
  ## Returns: average number of clock clicks
  ##
  proc averageClicksPerSecond {{rounds 5}} {
    set totalClicks 0
    for {set i 0} {$i < $rounds} {incr i} {
      incr totalClicks [countClicksPerSecond]
    }
    return [expr $totalClicks / $rounds]
  }

  ##
  ## Calibrate number of clock clicks in 1 second
  ##
  ## Args:    none
  ## Returns: nothing
  ##
  proc calibrateClicksPerSecond {} {
    variable NamespaceCurrent
    variable configData
    variable clicksPerSecond

    if {![info exists clicksPerSecond]} then {
      wpLog o * $NamespaceCurrent "Calibrating time routines -- this will take $configData(calrounds) seconds."
      set clicksPerSecond "[averageClicksPerSecond $configData(calrounds)] [clock seconds]"
    } elseif {[expr ([clock seconds] - [lindex $clicksPerSecond 1]) / 3600] >= $configData(recaltime)} then {
      wpLog o * $NamespaceCurrent "Recalibrating time routines -- this will take $configData(calrounds) seconds."
      set clicksPerSecond "[averageClicksPerSecond $configData(calrounds)] [clock seconds]"
    }
  }

  ##
  ## Return number of clock clicks in 1 second
  ##
  ## Args:    none
  ## Returns: clock clicks
  ##
  proc clicksPerSecond {} {
    variable clicksPerSecond

    if {![info exists clicksPerSecond]} then {
      calibrateClicksPerSecond
    }
    return [lindex $clicksPerSecond 0]
  }

  ##
  ## Recalibrate clock clicks per second if needed
  ##
  ## Args:    minute, hour, day, month, year
  ## Returns: 0
  ##
  proc time:hourly {minute hour day month year} {
    calibrateClicksPerSecond
    return 0
  }

  ##
  ## Init module specific things at load time
  ##
  ## Args:    action {load|unload}
  ## Returns: nothing
  ##
  proc InitTable {action} {
    variable NamespaceCurrent

    switch -exact -- $action {
      load {
        if {![info exists NamespaceCurrent]} then {
          set NamespaceCurrent [namespace current]
        }
      }
    }
    return
  }

  ##
  ## Bind procs to commands and events
  ##
  ## Args:    action {bind|unbind}, what {type|*}, args {...}
  ## Returns: nothing
  ##
  proc BindTable {action what args} {
    variable NamespaceCurrent

    if {![string compare * $what]} then {
      set what time
    }
    foreach type $what {
      switch -exact -- $type {
        time {
          $action time -|- "00 * * * *" ${NamespaceCurrent}::time:hourly
        }
      }
    }
    return
  }

  # Configuration data defaults
  array set configDataDefaults {
    calrounds 5
    recaltime 72
  }

  # Configuration data descriptions
  array set configDataDesc {
    calrounds "\[number\] rounds to run when checking clicks per second (at least 5)"
    recaltime "\[hours\] hours to wait before recalibrating clicks per second"
  }

} ;# namespace timetools

return
